# SPDX-FileCopyrightText: 2022-present Angus Hollands <goosey15@gmail.com>
#
# SPDX-License-Identifier: MIT
import pytest

from hatch_nodejs_version.metadata_source import NodeJSMetadataHook

TRIVIAL_PYPROJECT_CONTENTS = """
[build-backend]
requires = ["hatchling", "hatch-vcs"]
build-backend = "hatchling.build"
[project]
name = "my-app"
version = "0.0.0"
[tool.hatch.metadata.hooks.nodejs]
"""

TRIVIAL_PACKAGE_CONTENTS = """
{
  "name": "my-app",
  "version": "1.0.0",
  "description": "A terrible package",
  "keywords": [
    "what",
    "is",
    "the",
    "time"
  ],
  "homepage": "https://where-the-heart-is.com",
  "bugs": {
    "url": "https://www.send-help.com",
    "email": "funky@rider.com"
  },
  "license": "MIT",
  "author": {
    "name": "Alice Roberts",
    "email": "alice.roberts@bbc.lol"
  },
  "contributors": [
        {
        "name": "Isaac Newton",
        "email": "isaac.newton@apple.com"
      }
  ],
  "repository": {
    "type": "git",
    "url": "https://github.com/some/code.git"
  }
}
"""

TRIVIAL_EXPECTED_METADATA = {
    "license": "MIT",
    "urls": {
        "Bug Tracker": "https://www.send-help.com",
        "Repository": "https://github.com/some/code.git",
        "Homepage": "https://where-the-heart-is.com",
    },
    "authors": [
        {
            "name": "Alice Roberts",
            "email": "alice.roberts@bbc.lol",
        }
    ],
    "maintainers": [{"name": "Isaac Newton", "email": "isaac.newton@apple.com"}],
    "keywords": ["what", "is", "the", "time"],
    "description": "A terrible package",
    "name": "my-app",
}


class TestMetadata:
    @pytest.mark.parametrize(
        "alt_package_json",
        [None, "package-other.json"],
    )
    def test_all_metadata(self, project, alt_package_json):
        # Create a simple project
        package_json = "package.json" if alt_package_json is None else alt_package_json
        (project / "pyproject.toml").write_text(TRIVIAL_PYPROJECT_CONTENTS)
        (project / package_json).write_text(TRIVIAL_PACKAGE_CONTENTS)

        config = {} if alt_package_json is None else {"path": alt_package_json}
        metadata = {}
        metadata_source = NodeJSMetadataHook(project, config=config)
        metadata_source.update(metadata)

        assert metadata == TRIVIAL_EXPECTED_METADATA

    @pytest.mark.parametrize(
        "pyproject_field",
        [
            "urls",
            "description",
            "name",
            "keywords",
            "authors",
            "maintainers",
            "license",
        ],
    )
    def test_subset_metadata(self, project, pyproject_field):
        # Create a simple project
        (project / "pyproject.toml").write_text(TRIVIAL_PYPROJECT_CONTENTS)
        (project / "package.json").write_text(TRIVIAL_PACKAGE_CONTENTS)

        config = {"fields": [pyproject_field]}

        metadata = {}
        metadata_source = NodeJSMetadataHook(project, config=config)
        metadata_source.update(metadata)

        assert pyproject_field in metadata
        assert len(metadata) == len(config["fields"])
        assert metadata[pyproject_field] == TRIVIAL_EXPECTED_METADATA[pyproject_field]

    def test_contributors_as_maintainers(self, project):
        # Create a simple project
        (project / "pyproject.toml").write_text(TRIVIAL_PYPROJECT_CONTENTS)
        (project / "package.json").write_text(TRIVIAL_PACKAGE_CONTENTS)

        metadata = {}
        metadata_source = NodeJSMetadataHook(
            project, config={"contributors-as-maintainers": True}
        )
        metadata_source.update(metadata)

        assert metadata["authors"] == TRIVIAL_EXPECTED_METADATA["authors"]
        assert metadata["maintainers"] == TRIVIAL_EXPECTED_METADATA["maintainers"]

    def test_contributors_as_authors(self, project):
        # Create a simple project
        (project / "pyproject.toml").write_text(TRIVIAL_PYPROJECT_CONTENTS)
        (project / "package.json").write_text(TRIVIAL_PACKAGE_CONTENTS)

        metadata = {}
        metadata_source = NodeJSMetadataHook(
            project, config={"contributors-as-maintainers": False}
        )
        metadata_source.update(metadata)

        assert (
            metadata["authors"]
            == TRIVIAL_EXPECTED_METADATA["authors"]
            + TRIVIAL_EXPECTED_METADATA["maintainers"]
        )

    def test_labels(self, project):
        # Create a simple project
        (project / "pyproject.toml").write_text(TRIVIAL_PYPROJECT_CONTENTS)
        (project / "package.json").write_text(TRIVIAL_PACKAGE_CONTENTS)

        metadata = {}
        metadata_source = NodeJSMetadataHook(
            project,
            config={
                "repository-label": "the-repository",
                "bugs-label": "the-bug-tracker",
                "homepage-label": "the-homepage",
            },
        )
        metadata_source.update(metadata)

        urls = metadata["urls"]
        assert urls["the-repository"] == "https://github.com/some/code.git"
        assert urls["the-bug-tracker"] == "https://www.send-help.com"
        assert urls["the-homepage"] == "https://where-the-heart-is.com"



# SPDX-FileCopyrightText: 2022-present Angus Hollands <goosey15@gmail.com>
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import json
import os.path
import re
import urllib.parse
from typing import Any

from hatchling.metadata.plugin.interface import MetadataHookInterface

AUTHOR_PATTERN = r"^([^<(]+?)?[ \t]*(?:<([^>(]+?)>)?[ \t]*(?:\(([^)]+?)\)|$)"
REPOSITORY_PATTERN = r"^(?:(gist|bitbucket|gitlab|github):)?(.*?)$"
REPOSITORY_TABLE = {
    "gitlab": "https://gitlab.com",
    "github": "https://github.com",
    "gist": "https://gist.github.com",
    "bitbucket": "https://bitbucket.org",
}


class NodeJSMetadataHook(MetadataHookInterface):
    PLUGIN_NAME = "nodejs"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__path = None
        self.__fields = None
        self.__contributors_as_maintainers = None
        self.__homepage_label = None
        self.__bugs_label = None
        self.__repository_label = None

    @property
    def path(self) -> str:
        if self.__path is None:
            version_file = self.config.get("path", "package.json")
            if not isinstance(version_file, str):
                raise TypeError(
                    "Option `path` for metadata hook `{}` must be a string".format(
                        self.PLUGIN_NAME
                    )
                )

            self.__path = version_file

        return self.__path

    @property
    def fields(self) -> None | set[str]:
        if self.__fields is None:
            fields = self.config.get("fields", None)
            if fields is None:
                self.__fields = None
            else:
                if not (
                    isinstance(fields, list) and all(isinstance(f, str) for f in fields)
                ):
                    raise TypeError(
                        "Option `fields` for metadata hook `{}` "
                        "must be a list of strings".format(self.PLUGIN_NAME)
                    )
                self.__fields = set(fields)
        return self.__fields

    @property
    def contributors_as_maintainers(self) -> bool:
        if self.__contributors_as_maintainers is None:
            contributors_as_maintainers = self.config.get(
                "contributors-as-maintainers", True
            )
            if not isinstance(contributors_as_maintainers, bool):
                raise TypeError(
                    "Option `contributors-as-maintainers` for metadata hook `{}` "
                    "must be a boolean".format(self.PLUGIN_NAME)
                )
            self.__contributors_as_maintainers = contributors_as_maintainers
        return self.__contributors_as_maintainers

    @property
    def homepage_label(self) -> bool:
        if self.__homepage_label is None:
            homepage_label = self.config.get("homepage-label", "Homepage")

            if not isinstance(homepage_label, str):
                raise TypeError(
                    "Option `homepage-label` for metadata hook `{}` "
                    "must be a string".format(self.PLUGIN_NAME)
                )
            self.__homepage_label = homepage_label
        return self.__homepage_label

    @property
    def bugs_label(self) -> bool:
        if self.__bugs_label is None:
            bug_tracker_label = self.config.get("bugs-label", "Bug Tracker")

            if not isinstance(bug_tracker_label, str):
                raise TypeError(
                    "Option `bugs-label` for metadata hook `{}` "
                    "must be a string".format(self.PLUGIN_NAME)
                )
            self.__bugs_label = bug_tracker_label
        return self.__bugs_label

    @property
    def repository_label(self) -> bool:
        if self.__repository_label is None:
            bug_tracker_label = self.config.get("repository-label", "Repository")

            if not isinstance(bug_tracker_label, str):
                raise TypeError(
                    "Option `repository-label` for metadata hook `{}` "
                    "must be a string".format(self.PLUGIN_NAME)
                )
            self.__repository_label = bug_tracker_label
        return self.__repository_label

    def load_package_data(self):
        path = os.path.normpath(os.path.join(self.root, self.path))
        if not os.path.isfile(path):
            raise OSError(f"file does not exist: {self.path}")

        with open(path, "r", encoding="utf-8") as f:
            return json.load(f)

    def _parse_bugs(self, bugs: str | dict[str, str]) -> str | None:
        if isinstance(bugs, str):
            return bugs

        if "url" not in bugs:
            return None

        return bugs["url"]

    def _parse_person(self, person: dict[str, str]) -> dict[str, str]:
        if {"url", "email"} & person.keys():
            result = {"name": person["name"]}
            if "email" in person:
                result["email"] = person["email"]
        else:
            match = re.match(AUTHOR_PATTERN, person["name"])
            if match is None:
                raise ValueError(f"Invalid author name: {person['name']}")
            name, email, _ = match.groups()
            result = {"name": name}
            if email is not None:
                result["email"] = email

        return result

    def _parse_repository(self, repository: str | dict[str, str]) -> str:
        if isinstance(repository, str):
            match = re.match(REPOSITORY_PATTERN, repository)
            if match is None:
                raise ValueError(f"Invalid repository string: {repository}")
            kind, identifier = match.groups()
            if kind is None:
                kind = "github"
            return urllib.parse.urljoin(REPOSITORY_TABLE[kind], identifier)

        return repository["url"]

    def update(self, metadata: dict[str, Any]):
        package = self.load_package_data()

        new_metadata = {"name": package["name"]}

        authors = None
        maintainers = None

        if "author" in package:
            authors = [self._parse_person(package["author"])]

        if "contributors" in package:
            contributors = [self._parse_person(p) for p in package["contributors"]]
            if self.contributors_as_maintainers:
                maintainers = contributors
            else:
                authors = [*(authors or []), *contributors]

        if authors is not None:
            new_metadata["authors"] = authors

        if maintainers is not None:
            new_metadata["maintainers"] = maintainers

        if "keywords" in package:
            new_metadata["keywords"] = package["keywords"]

        if "description" in package:
            new_metadata["description"] = package["description"]

        if "license" in package:
            new_metadata["license"] = package["license"]

        # Construct URLs
        urls = {}
        if "homepage" in package:
            urls[self.homepage_label] = package["homepage"]
        if "bugs" in package:
            bugs_url = self._parse_bugs(package["bugs"])
            if bugs_url is not None:
                urls[self.bugs_label] = bugs_url
        if "repository" in package:
            urls[self.repository_label] = self._parse_repository(package["repository"])

        # Write URLs
        if urls:
            new_metadata["urls"] = urls

        # Only use required metadata
        metadata.update(
            {
                k: v
                for k, v in new_metadata.items()
                if (self.fields is None or k in self.fields)
            }
        )



# SPDX-FileCopyrightText: 2022-present Angus Hollands <goosey15@gmail.com>
# SPDX-FileCopyrightText: 2022-present Ofek Lev <oss@ofek.dev>
#
# SPDX-License-Identifier: MIT
import errno
import os
import shutil
import stat
import tempfile
from contextlib import contextmanager
import pathlib

import pytest


def touch_file(path):
    with open(path, "a"):
        os.utime(path, None)


def create_file(path, contents):
    with open(path, "w") as f:
        f.write(contents)


@contextmanager
def create_project(directory):
    project_dir = directory / "my-app"
    os.mkdir(project_dir)

    package_dir = project_dir / "my_app"
    os.mkdir(package_dir)

    touch_file(package_dir / "__init__.py")
    touch_file(package_dir / "foo.py")
    touch_file(package_dir / "bar.py")
    touch_file(package_dir / "baz.py")

    origin = os.getcwd()
    os.chdir(project_dir)
    try:
        yield project_dir
    finally:
        os.chdir(origin)


def handle_remove_readonly(func, path, exc):  # no cov
    # PermissionError: [WinError 5] Access is denied: '...\\.git\\...'
    if func in (os.rmdir, os.remove, os.unlink) and exc[1].errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
        func(path)
    else:
        raise


@pytest.fixture
def temp_dir():
    directory = tempfile.mkdtemp()
    try:
        directory = pathlib.Path(os.path.realpath(directory))
        yield directory
    finally:
        shutil.rmtree(directory, ignore_errors=False, onerror=handle_remove_readonly)


@pytest.fixture
def project(temp_dir):
    with create_project(temp_dir) as project:
        yield project



# SPDX-FileCopyrightText: 2022-present Angus Hollands <goosey15@gmail.com>
#
# SPDX-License-Identifier: MIT
import json

import pytest

from hatch_nodejs_version.version_source import NodeJSVersionSource

GOOD_NODE_PYTHON_VERSIONS = [
    ("1.4.5", "1.4.5"),
    ("1.4.5-a0", "1.4.5a0"),
    ("1.4.5-a", "1.4.5a"),
    ("1.4.5-b0", "1.4.5b0"),
    ("1.4.5-c1", "1.4.5c1"),
    ("1.4.5-rc0", "1.4.5rc0"),
    ("1.4.5-alpha0", "1.4.5alpha0"),
    ("1.4.5-beta0", "1.4.5beta0"),
    ("1.4.5-pre9", "1.4.5pre9"),
    ("1.4.5-preview0", "1.4.5preview0"),
    ("1.4.5-preview0+build1.0.0", "1.4.5preview0+build1.0.0"),
    ("1.4.5-preview0+build-1.0.0", "1.4.5preview0+build-1.0.0"),
    ("1.4.5-preview0+good-1_0.0", "1.4.5preview0+good-1_0.0"),
]


class TestVersion:
    @pytest.mark.parametrize(
        "python_version",
        [
            "1.4",
            "1.4.5ab",
            "1.4.5-c0.smoke2",
            "1.4.5rc.post1@dev2",
            "1.4.5rc0.post1+-bad",
            "1.4.5rc0.post1+bad_",
        ],
    )
    def test_parse_python_incorrect(self, python_version):
        with pytest.raises(ValueError, match=".* did not match regex"):
            NodeJSVersionSource.python_version_to_node(python_version)

    @pytest.mark.parametrize(
        "node_version",
        [
            "1.4",
            "1.4.5a0",
            "1.4.5-c0.post1",
            "1.4.5-rc0.post1.dev2",
            "1.4.5-rc0.post1+-bad",
            "1.4.5-rc0.post1+bad_",
        ],
    )
    def test_parse_node_incorrect(self, node_version):
        with pytest.raises(ValueError, match=".* did not match regex"):
            NodeJSVersionSource.node_version_to_python(node_version)

    @pytest.mark.parametrize(
        "node_version, python_version",
        GOOD_NODE_PYTHON_VERSIONS,
    )
    @pytest.mark.parametrize(
        "alt_package_json",
        [None, "package-other.json"],
    )
    def test_version_from_package(
        self, project, node_version, python_version, alt_package_json
    ):
        # Create a simple project
        (project / "pyproject.toml").write_text(
            """
[build-system]
requires = ["hatchling", "hatch-vcs"]
build-backend = "hatchling.build"
[project]
name = "my-app"
dynamic = ["version"]
[tool.hatch.version]
source = "nodejs"
 """
        )
        package_json = "package.json" if alt_package_json is None else alt_package_json
        (project / package_json).write_text(
            f"""
{{
  "name": "my-app",
  "version": "{node_version}"
}}
"""
        )
        config = {} if alt_package_json is None else {"path": alt_package_json}
        version_source = NodeJSVersionSource(project, config=config)
        data = version_source.get_version_data()
        assert data["version"] == python_version

    @pytest.mark.parametrize(
        "node_version, python_version",
        GOOD_NODE_PYTHON_VERSIONS,
    )
    @pytest.mark.parametrize(
        "alt_package_json",
        [None, "package-other.json"],
    )
    def test_version_to_package(
        self, project, node_version, python_version, alt_package_json
    ):
        package_json = "package.json" if alt_package_json is None else alt_package_json
        (project / "pyproject.toml").write_text(
            """
[build-system]
requires = ["hatchling", "hatch-vcs"]
build-backend = "hatchling.build"
[project]
name = "my-app"
dynamic = ["version"]
[tool.hatch.version]
source = "nodejs"
 """
        )
        (project / package_json).write_text(
            """
{
  "name": "my-app",
  "version": "0.0.0"
}
"""
        )
        config = {} if alt_package_json is None else {"path": alt_package_json}
        version_source = NodeJSVersionSource(project, config=config)
        version_data = version_source.get_version_data()
        version_source.set_version(python_version, version_data)

        written_package = json.loads((project / package_json).read_text())
        assert written_package["version"] == node_version



# SPDX-FileCopyrightText: 2022-present Angus Hollands <goosey15@gmail.com>
#
# SPDX-License-Identifier: MIT
from hatchling.plugin import hookimpl

from .version_source import NodeJSVersionSource
from .metadata_source import NodeJSMetadataHook


@hookimpl
def hatch_register_version_source():
    return NodeJSVersionSource


@hookimpl
def hatch_register_metadata_hook():
    return NodeJSMetadataHook
